package ps6.graph;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.List;
import java.util.Set;



/**
 * <code>DefaultMutableDirectedLabeledMultiGraph</code> is the default implementation
 * of @link{ps6.graph.MutableDirectedLabeledMultiGraph}.
 *
 * <code>DefaultMutableDirectedLabeledMultiGraph</code> is not synchronized.
 *
 * @author mbolin
 * @author cwo (modified for Java 1.5)
 */
public class DefaultMutableDirectedLabeledMultiGraph<T> implements MutableDirectedLabeledMultiGraph<T> {

    /*
     * All implementations of Map used in this implementation
     * must support null keys.
     * 
     * All implementations of Set used in this implementation
     * must support null values.
     */
    
	/*
	 * AF:
	 * nodes = nodes
	 * edges = {e: Edge | (some es: outgoingEdgeMap.values | e in es)}
	 * 
	 */

    /*
     * RI:
     * (1) The keys in outgoingEdgeMap should be the same as those in incomingEdgeMap
     * (2) Every value in outgoingEdgeMap and in incomingEdgeMap is a non-null Map
     * (3) The sum of the size of the mappings over all the values in outgoingEdgeMap
     *  is the sum of the size of the mappings over all the values in incomingEdgeMap
     * (4) The sums listed in (3) are both equal to edgeCount
     * (5) [NOT TESTED] every incoming edge is also an outgoing edge with the same label
     * (6) [NOT TESTED] every outgoing edge is also an incoming edge with the same label
     */
    private static final boolean CHECKREP_ENABLED = true;

    private void checkRep() {
        if (!CHECKREP_ENABLED) {
            return;
        }

        // check that keysets are equal
        if (!outgoingEdgeMap.keySet().equals(incomingEdgeMap.keySet())) {
            throw new RuntimeException(
                    "outgoing and incoming edge maps have different keySets");
        }

        // check number of incoming edges is correct
        int incomingEdges = 0;
        for (Iterator iter = incomingEdgeMap.values().iterator(); iter.hasNext();) {
            incomingEdges += ((Set) iter.next()).size();
        }
        if (incomingEdges != edgeCount) {
            throw new RuntimeException(
                    "number of incoming edges does not agree with edgeCount");
        }

        // check number of outgoing edges is correct
        int outgoingEdges = 0;
        for (Iterator iter = outgoingEdgeMap.values().iterator(); iter.hasNext();) {
            outgoingEdges += ((Set)iter.next()).size();
        }
        if (outgoingEdges != edgeCount) {
            throw new RuntimeException(
                    "number of outgoing edges does not agree with edgeCount");
        }
    }

    /**
     * The key is a node in the graph, and
     * the value is a mapping from labels to outgoing edges for the node.
     */
    private final Map<T ,Set<Edge<T> > > outgoingEdgeMap = new HashMap<T, Set<Edge<T> > >();

    /**
     * The key is a node in the graph, and
     * the value is a mapping from labels to outgoing edges for the node.
     */
    private final Map<T, Set<Edge<T> > > incomingEdgeMap = new HashMap<T, Set<Edge<T> > >();

    private final Set<T> nodes = outgoingEdgeMap.keySet();

    /** The number of edges in the graph. */
    private int edgeCount = 0;

    /**
     * Creates an empty graph with no nodes or edges.
     */
    public DefaultMutableDirectedLabeledMultiGraph() {
        checkRep();
    }

    /* (non-Javadoc)
     * @see ps6.graph.MutableDirectedLabeledMultiGraph#addNode(java.lang.Object)
     */
    public boolean addNode(T obj) {
        checkRep();
        try {
            if (nodes.contains(obj)) {
                // graph already contains obj
                return false;
            }
            Set<Edge<T> > incomingEdges = new HashSet<Edge<T> >(); 
            incomingEdgeMap.put(obj, incomingEdges);
            Set<Edge<T> > outgoingEdges = new HashSet<Edge<T> >(); 
            outgoingEdgeMap.put(obj, outgoingEdges);
            return true;
        } finally {
            checkRep();
        }
    }
    
    public Iterator<T> getNodes() {
        return Collections.unmodifiableSet(nodes).iterator();
    }
    
    /* (non-Javadoc)
     * @see ps6.graph.MutableDirectedLabeledMultiGraph#addEdge(ps6.graph.DefaultEdge)
     */
    public boolean addEdge(Edge<T> e) {
        checkRep();
        try {
            Set< Edge<T> > outgoingEdges = outgoingEdgeMap.get(e.getSource());
            if (outgoingEdges == null) { return false; }
            if (! outgoingEdges.add(e)) { return false; }
            Set<Edge<T>> incomingEdges = incomingEdgeMap.get(e.getSink());
            incomingEdges.add(new DefaultEdge<T>(e.getSource(), e.getSink(), e.getLabel()));
            edgeCount++;
            return true;
        } finally {
            checkRep();
        }
    }
    
    public Iterator<Edge<T> > getEdges() {
        // consider making this a constant time operation
        // by maintaining a set of allEdges
        //
        // debatable whether this operation is used often
        // enough to merit the extra storage overhead
        Set<Edge<T> > allEdges = new HashSet<Edge<T> >(); 
        for (Iterator<Set<Edge<T> > > iter = outgoingEdgeMap.values().iterator(); iter.hasNext(); ) {
            Set<Edge<T> > edges = iter.next();
            allEdges.addAll(edges);
        }
        return Collections.unmodifiableSet(allEdges).iterator();
    }

    /* (non-Javadoc)
     * @see ps6.graph.MutableDirectedLabeledMultiGraph#removeNode(java.lang.Object)
     */
    public boolean removeNode(T obj) {
        checkRep();
        try {
            if (!containsNode(obj)) {
                return false;
            }
            // change this algorithm so it runs in constant time
            Set<Edge<T> > incomingEdges = incomingEdgeMap.remove(obj);
            Set<Edge<T> > outgoingEdges = outgoingEdgeMap.remove(obj);
            edgeCount -= (incomingEdges.size() + outgoingEdges.size());
            
            for (Iterator<Edge<T> > iter = incomingEdges.iterator(); iter.hasNext(); ) {
                Edge<T> e = iter.next();
                Set<Edge<T>> outgoing = outgoingEdgeMap.get(e.getSource());
                outgoing.remove(e);
            }
            
            for (Iterator<Edge<T> > iter = outgoingEdges.iterator(); iter.hasNext(); ) {
                Edge<T> e = iter.next();
                Set<Edge<T> > incoming = incomingEdgeMap.get(e.getSink());
                incoming.remove(e);
            }
            return true;
        } finally {
            checkRep();
        }
    }
    
    /* (non-Javadoc)
     * @see ps6.graph.MutableDirectedLabeledMultiGraph#removeEdge(ps6.graph.DefaultEdge)
     */
    public boolean removeEdge(Edge<T> e) {
        checkRep();
        try {
            if (!containsEdge(e)) {
                return false;
            }
            Set outgoingEdges = (Set) outgoingEdgeMap.get(e.getSource());
            outgoingEdges.remove(e);
            Set incomingEdges = (Set) incomingEdgeMap.get(e.getSink());
            incomingEdges.remove(e);
            edgeCount--;
            return true;
        } finally {
            checkRep();
        }
    }

    /* (non-Javadoc)
     * @see ps6.graph.MultiGraph#containsNode(java.lang.Object)
     */
    public boolean containsNode(T obj) {
        checkRep();
        return nodes.contains(obj);
    }
    
    /* (non-Javadoc)
     * @see ps6.graph.MultiGraph#containsEdge(java.lang.Object, java.lang.Object, java.lang.Object)
     */
    public boolean containsEdge(Edge<T> edge) {
    	T from = edge.getSource();
    	Object label = edge.getLabel();
    	T to = edge.getSink();
        checkRep();
        Set<Edge<T> > outgoingEdges = outgoingEdgeMap.get(from);
        if (outgoingEdges == null) { return false; }
        
        // abstract this out and make it a protected method
        // that uses EdgeSet for better performance
        // and performs the operation in constant time
        for (Iterator<Edge<T> > iter = outgoingEdges.iterator(); iter.hasNext(); ) {
            Edge<T> e = iter.next();
            Object eLabel = e.getLabel();
            if (!GraphUtils.eq(label, eLabel)) { continue; }
            T eTo = e.getSink();
            if (GraphUtils.eq(to, eTo)) { return true; }
        }
        return false;
    }
    
    public int getOutgoingEdgeCount(T obj) {
        Set<Edge<T> > edges = outgoingEdgeMap.get(obj);
        return (edges == null) ? 0 : edges.size();
    }
    
    /* (non-Javadoc)
     * @see ps6.graph.MultiGraph#getOutgoingEdges(java.lang.Object)
     */
    public Iterator<Edge<T> > getOutgoingEdges(T obj) {
        checkRep();
        if (!containsNode(obj)) {
            List<Edge<T> > myempty = Collections.emptyList();
            return myempty.iterator();
        }
        return Collections.unmodifiableSet(outgoingEdgeMap.get(obj)).iterator();
    }

    public int getIncomingEdgeCount(T obj) {
        Set<Edge<T> > edges = incomingEdgeMap.get(obj);
        return (edges == null) ? 0 : edges.size();
    }
    
    /* (non-Javadoc)
     * @see ps6.graph.MultiGraph#getIncomingEdges(java.lang.Object)
     */
    public Iterator<Edge<T> > getIncomingEdges(T obj) {
        checkRep();
        if (!containsNode(obj)) {
            List<Edge<T> > myempty = Collections.emptyList();
            return myempty.iterator();
        }
        return Collections.unmodifiableSet(incomingEdgeMap.get(obj)).iterator();
    }

    /* (non-Javadoc)
     * @see ps6.graph.MultiGraph#getNodeCount()
     */
    public int getNodeCount() {
        checkRep();
        return nodes.size();
    }

    /* (non-Javadoc)
     * @see ps6.graph.MultiGraph#getEdgeCount()
     */
    public int getEdgeCount() {
        checkRep();
        return edgeCount;
    }

}
