package ps6.graph;

import java.util.Iterator;

/**
 * <code>DirectedLabeledMultiGraph</code> represents a directed labeled multi-graph.
 * <code>DirectedLabeledMultiGraph</code> is immutable.
 * <p/>
 * Though the structure of the graph is preserved by this interface
 * (i.e., nodes and edges may not be removed or deleted),
 * the nodes and edge labels in the graph may be references to
 * the elements in the graph rather than references to copies
 * of those elements, and these references may allow a client
 * of this interface to mutate an element in the graph.
 * When this is the case, these elements
 * may not be mutated in such a way that affects equals comparisons
 * while they are elements in the graph.
 * (This is analogous to the restrictions on mutating keys in
 * the {@link java.util.Map} interface.)
 * <p/>
 * <b>Note</b>: Throughout this API, the notation <code>e1.equals(e2)</code> is used to mean
 * <code>(e1==null ? e2==null : e1.<b>equals</b>(e2))</code> where the second
 * <code><b>equals</b></code> is the method that tests for object equality.
 * 
 * @specfield (set Object) nodes // set of unique elements that constitute the nodes in the graph
 * @specfield (set Edge) edges // set of 3-tuples representing edges where each tuple has
 *                           a source node, a sink node, and a label where source
 *                           and sink are elements of nodes
 * abstract invariant:
 * 	edges.source in nodes && edges.sink in nodes
 * 	all n, n': nodes | n != n' => !n.equals(n')
 * 	all e, e': edges | e != e' => !e.equals(e')
 * @see java.util.Map
 * @see ps6.graph.GraphUtils#eq(Object,Object)
 * 
 * @author mbolin
 * @author cwo (modified for Java 1.5) 
 */
public interface DirectedLabeledMultiGraph<T> {
    
    /**
     * Indicates if the node is contained in this graph.
     * 
     * @returns <code>true</code> if <code>nodes</code> contains a node <code>n</code>
     * such that <code>n.equals(obj)</code>,
     * <code>false</code> otherwise
     */
    public boolean containsNode(T obj);
    
    /** 
     * Indicates if the edge is contained in this graph.
     * 
     * @returns <code>true</code> if <code>edges</code> contains an edge <code>e</code>
     * such that <code>e.equals(edge)</code>,
     * <code>false</code> otherwise
     */
    public boolean containsEdge(Edge<T> edge);
    
    /**
     * Returns an <code>Iterator</code> over the nodes in this graph.
     * 
     * @returns an <code>Iterator&lt;Object&gt;</code> where each element
     * yielded by the <code>Iterator</code> is a node in the graph.
     * The <code>remove()</code> method of this <code>Iterator</code>
     * is not supported.
     * 
     * @see java.util.Iterator#remove()
     */
    public Iterator<T> getNodes();

    /**
     * Returns an <code>Iterator</code> over the edges in this graph.
     * 
     * @returns an <code>Iterator&lt;Edge&gt;</code> where each element
     * yielded by the <code>Iterator</code> is a non-null <code>Edge</code> in the graph.
     * The <code>remove()</code> method of this <code>Iterator</code>
     * is not supported.
     * 
     * @see java.util.Iterator#remove()
     */    
    public Iterator<Edge<T>> getEdges();

    /**
     * Returns the number of nodes in the graph.
     * 
     * @returns size of <code>nodes</code>
     */
    public int getNodeCount();
    
    /**
     * Returns the number of edges in the graph.
     * 
     * @returns size of <code>edges</code>
     */
    public int getEdgeCount();    
    
    /**
     * Returns an <code>Iterator</code> over the set of outgoing edges
     * from <code>obj</code>.
     * 
     * An outgoing edge from <code>obj</code> is defined as any edge <code>e</code> 
     * in the graph where <code>e.getSource().equals(obj)</code>
     * 
     * @returns an <code>Iterator&lt;Edge&gt;</code> where each element
     * of the <code>Iterator</code> is a non-null <code>Edge e</code> where
     * <code>e.getSource().equals(obj)</code>.
     * The <code>remove()</code> method of this <code>Iterator</code>
     * is not supported.
     * 
     * If <code>obj</code> is not in the graph, then an empty Iterator will be returned.
     * Thus, <code>getOutgoingEdgeCount(obj)</code> should agree with the number
     * of elements in the <code>Iterator</code>
     * returned by <code>getOutgoingEdges(obj)</code>.
     * 
     * @see #getOutgoingEdgeCount(Object)
     */
    public Iterator<Edge<T> > getOutgoingEdges(T obj);
    
    /**
     * Returns an <code>Iterator</code> over the set of incoming edges
     * to <code>obj</code>.
     * 
     * An incoming edge to <code>obj</code> is defined as any edge <code>e</code> 
     * in the graph where <code>e.getSink().equals(obj)</code>
     *  
     * @returns an <code>Iterator&lt;Edge&gt;</code> where each element
     * of the <code>Iterator</code> is a non-null <code>Edge e</code> where
     * <code>e.getSink().equals(obj)</code>.
     * The <code>remove()</code> method of this <code>Iterator</code>
     * is not supported.
     * 
     * If <code>obj</code> is not in the graph, then an empty Iterator will be returned.
     * Thus, <code>getIncomingEdgeCount(obj)</code> should agree with the number
     * of elements in the <code>Iterator</code>
     * returned by <code>getIncomingEdges(obj)</code>.
     * 
     * @see #getIncomingEdgeCount(Object)
     */    
    public Iterator<Edge<T> > getIncomingEdges(T obj);
    
    /**
     * Returns the number of outgoing edges from <code>obj</code>.
     * 
     * An outgoing edge from <code>obj</code> is defined as any edge <code>e</code>
     * in the graph where <code>e.getSource().equals(obj)</code>
     * 
     * @returns the number of outgoing edges from <code>obj</code>,
     * 0 if <code>obj</code> is not in the graph
     * 
     * @see #getOutgoingEdges(Object)
     */
    public int getOutgoingEdgeCount(T obj);

    /**
     * Returns the number of incoming edges to <code>obj</code>.
     * 
     * An incoming edge to <code>obj</code> is defined as any edge <code>e</code>
     * in the graph where <code>e.getSink().equals(obj)</code>
     * 
     * @returns the number of incoming edges from <code>obj</code>,
     * 0 if <code>obj</code> is not in the graph
     * 
     * @see #getIncomingEdges(Object)
     */
    public int getIncomingEdgeCount(T obj);
    
}
